
KX=PolynomialRing(QQ,['X1','X2','X3','X4','X5','X6','X7','X8','r','s'])
K=PolynomialRing(QQ,['r','s'])
KX.inject_variables()
K.inject_variables()

# cosines theorem
def kashi(x,y,z):
    return ((y+x)^2+(y+z)^2-(x+z)^2)/2/(y+x)/(y+z)

# cosines theorem to get cosines as rational fractions in r and s
cosinus=[
kashi(1,s,1),kashi(1,s,r),kashi(1,s,s),kashi(r,s,r),kashi(r,s,s),kashi(s,s,s),
kashi(1,r,1),kashi(1,r,r),kashi(1,r,s),kashi(r,r,r),kashi(r,r,s),kashi(s,r,s),
kashi(1,1,1),kashi(1,1,r),kashi(1,1,s),kashi(r,1,r),kashi(r,1,s),kashi(s,1,s)
]

# sines using sqrt(1-cos^2) and auxiliary variables Xi to avoid square root
sinus=[
2*X1/(s+1)^2, 2*X2/(s+1)/(r+s), X3/(s+1), 2*r*X4/(r+s)^2, X5/(r+s), X6/2,
2*X7/(r+1)^2, X8/(r+1), 2*X2/(r+1)/(r+s), X6/2, X4/(r+s), 2*s*X5/(r+s)^2,
X6/2, X7/(r+1), X1/(s+1), 2*r*X8/(r+1)^2, 2*X2/(r+1)/(s+1), 2*s*X3/(s+1)^2
]

# definition of the auxiliary variables
dico=dict({X1:s*(s+2),X2:r*s*(r+s+1),X3:2*s+1,X4:s*(2*r+s),X5:r*(r+2*s),X6:3,X7:r*(r+2),X8:2*r+1})

# a factor of a polynomial is useless if it has no real root 0<s<r<1
# here is a small heuristic to remove some useless factors (to eventually get a simpler polynomial) : all the coefficients avec the same sign or some frequence cases
def useless(f):
    return len(set(sign(i) for i in f.coefficients()))==1 or f in [s-1,1-s,r-1,1-r,r-s,s-r]

# remove the auxiliary variables (each variable roughly double the equation degree)
def elimine(eqX):
    try:
        # remove useless factors and multiplicities
        l=[f[0] for f in factor(eqX) if not set(f[0].variables()).issubset({r,s}) or not useless(f[0])]
#        if len(l)>1:
#            return prod([elimine(f) for f in l])
        eqX=prod(l)
        # choose an Xi and remove it
        X=set(eqX.variables()).difference({r,s})
        Xi=X.pop()
        A=eqX.coefficient(Xi)
        B=eqX-A*Xi
        eqX=A^2*dico.get(Xi)-B^2 # the most time-consuming step
        # remove any Xj^2 that could have be created
        for Xj in X:
            eqX+=eqX.coefficient(Xj^2)*(dico.get(Xj)-Xj^2)
        return elimine(eqX)
    except KeyError: # no more Xi to remove
        return eqX

# polynomial with auxiliary variables associated with the vector angle k
# tc=0,1,2 stands for a small/medium/large corona
def equationX(k,tc):
    (a,b,c,d,e,f)=SR.var('a b c d e f')
    v=(a,b,c,d,e,f)
    l=[cos(i) for i in v]+[sin(i) for i in v]
    eqX=KX(-1) # cos(2pi) on right-hand side of the equation
    if len([1 for i in k if i<>0])==1: # special case because operands(cos(a)) returns a and not cos(a)
        ops=[cos(v[i]) for i in range(6) if k[i]<>0]
    else: # usual case
        ops=cos(sum(v[i]*sign(k[i]) for i in range(6))).trig_expand().operands()
    for o in ops:
        t=KX(1)
        for ft in o.factor_list():
            if ft[0]==-1:
                t=-t
            else:
                i=l.index(ft[0])
                if i<6: #cos
                    t*=chebyshev_T(k[i],cosinus[6*tc+i])
                else: #sin
                    t*=chebyshev_U(k[i-6]-1,cosinus[6*tc+i-6])*sinus[6*tc+i-6]
        eqX+=t
    return KX(prod([f[0] for f in factor(numerator(eqX)) if not useless(f[0])]))
#    return KX(numerator(eqX))

# polynomial in r and s associated with the vector angle k
def equation(k,tc):
    return K(elimine(equationX(k,tc)))

#################
## vérification ##

# check that the equation eqX with auxiliary variables Xi is satisfied by the algebraic values rs=(r,s)
def elimine2(eqX,rs):
    try:
        # remove useless factors and multiplicities
        l=[f[0] for f in factor(eqX) if not useless(f[0])]
#        if len(l)>1: # factorisation possible (tjrs facteur sans Xi ?)
#            return sum([elimine2(f,rs) for f in l])>0
        eqX=KX(prod(l))
        # choose an Xi and remove it
        X=set(eqX.variables()).difference({r,s})
        Xi=X.pop()
        A=eqX.coefficient(Xi)
        B=eqX-A*Xi
        # before multiplying by AXi-B, check it is non-zero in interval arithmetic
        C=A*Xi-B
        C=C.substitute({Xi:sqrt(SR(dico.get(Xi))) for Xi in set(eqX.variables()).difference({r,s})})
        C=RIF(C.substitute({r:RIF(rs[0]),s:RIF(rs[1])}))
        # if interval arithmetic does not ensures that it is non-zero, raise a Warning (will be checked exactly elsewhere)
        if C.contains_zero():
            raise Warning
        eqX=A^2*dico.get(Xi)-B^2
        # remove any Xj^2 that could have be created
        for Xj in X:
            eqX+=eqX.coefficient(Xj^2)*(dico.get(Xj)-Xj^2)
        return elimine2(eqX,rs)
    except KeyError: # no more Xi to remove: check exactly the equation
        a=K(eqX).substitute({r:rs[0],s:rs[1]})
        a.exactify()
        return a==0

# check exactly whether the corona k of type tc is ok with values rr and ss of r and s
def check_corona_exact(k,tc,rr,ss):
    eqX=equationX(k,tc)
    try:
        return elimine2(eqX,(rr,ss))
    except Warning: # interval arithmetic does not suffices -> substitutes algebraic values of r and s in Xi (more time-consuming)
        eq=eqX.substitute({Xi:sqrt(SR(dico.get(Xi))) for Xi in set(eqX.variables()).difference({r,s})})
        z=AA(eq.substitute(r=rr,s=ss))
        return z.exactify()==0
    return true

